#include "controlframe.h"
#include "renderFrame.h"
#include "logframe.h"
#include "TriMesh.h"

namespace AMT
{

enum
{
    ID_QUIT  = wxID_EXIT,
    ID_OPENOBJ = 101
};

BEGIN_EVENT_TABLE(CControlFrame,wxFrame)
  EVT_MENU    (ID_QUIT,  CControlFrame::OnQuit)
  EVT_MENU    (ID_OPENOBJ,  CControlFrame::OnLoadObj)
  EVT_CLOSE	  (CControlFrame::OnClose) 
END_EVENT_TABLE()

CControlFrame::CControlFrame( CLOGFrame *lf, const wxString& title, const wxPoint& pos,
        const wxSize& size, long style )
		: wxFrame( NULL, wxID_ANY, title, pos, size, style ), m_lf(lf)
{
	wxMenuBar *menu_bar = new wxMenuBar();

	wxMenu *menuFile = new wxMenu;
	menuFile->Append( ID_OPENOBJ, _T("&Load OBJ file...\tCtrl-L"));
	menuFile->AppendSeparator();
	menuFile->Append( ID_QUIT, _T("E&xit\tCtrl-Q"));
	menu_bar->Append(menuFile, _T("&File"));

	SetMenuBar( menu_bar );
}

void CControlFrame::OnClose( wxCloseEvent& event )
{
	for ( int i = 0; i < CRenderFrame::renderFrameList.size(); i++ )
		CRenderFrame::renderFrameList[i]->Close(true);
	m_lf->Close(true);
	Destroy();
}

void CControlFrame::OnQuit(wxCommandEvent &WXUNUSED(event))
{
	Close(true);
}

void CControlFrame::OnLoadObj(wxCommandEvent &WXUNUSED(event))
{
    wxString filename = wxFileSelector(_T("Select OBJ file"));
    if ( !filename )
        return;

	if ( CRenderFrame::lastFocusedFrame != NULL )
	{
		CTriMesh *triMesh = new CTriMesh( filename.c_str() );
		CRenderFrame::lastFocusedFrame->AddRenderObject( triMesh );
		CRenderFrame::lastFocusedFrame->selectedObject = (CRenderObject*) triMesh;
		CRenderFrame::lastFocusedFrame->Refresh(false);
	}
	else
		wxLogError( "Error: There is no Render Window on Focus!" );
}

}