#include <GL\glew.h>
#include "TriMesh.h"
#include "wx\wx.h"
#include <stdio.h>
#include <string.h>
#include <GL\gl.h>
#include "glm.h"
extern "C"
{
#include "trackball.h"
}

namespace AMT
{

static void
glmCross(float* u, float* v, float* n)
{
    assert(u); assert(v); assert(n);
    
    n[0] = u[1]*v[2] - u[2]*v[1];
    n[1] = u[2]*v[0] - u[0]*v[2];
    n[2] = u[0]*v[1] - u[1]*v[0];
}

static void
glmMinus(float* u, float* v, float* n)
{
    assert(u); assert(v); assert(n);
    
    n[0] = u[0] - v[0];
    n[1] = u[1] - v[1];
    n[2] = u[2] - v[2];
}

static void
glmAssign(float* f, float* t)
{
    assert(f); assert(t);
    
    t[0] = f[0];
    t[1] = f[1];
    t[2] = f[2];
}

static void
glmNormalize(float* v)
{
    float l;
    assert(v);
    l = 1.0f / (float)sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
    v[0] *= l;
    v[1] *= l;
    v[2] *= l;
}

CTriMesh::CTriMesh()
	: CRenderObject( true, 1.0f ), numVertice(0), numFaces(0), vertice(NULL), numNormals(0), numTextureCoords(0), \
	  vNormals(NULL),  textureCoords(NULL), materials(NULL), m_rendermode(WIREFRAME)
{
	m_IsTriMesh = true;
	m_InverseNormal = false;
}

CTriMesh::CTriMesh( const char *fileName )
	: CRenderObject( true, 1.0f ), numVertice(0), numFaces(0), vertice(NULL), numNormals(0), numTextureCoords(0), \
	  vNormals(NULL), textureCoords(NULL), materials(NULL), m_rendermode(WIREFRAME)
{
	m_IsTriMesh = true;
	m_InverseNormal = false;
	loadFile( fileName );
}
/*
void CTriMesh::prepareVBO()
{
	glGenBuffers(1, &vertexbufferID);
	glBindBuffer(GL_ARRAY_BUFFER, vertexbufferID);
	glBufferData(GL_ARRAY_BUFFER, numVertice*3*sizeof(float), vertice, GL_STATIC_DRAW);

	glGenBuffers(1, &normalbufferID);
	glBindBuffer(GL_ARRAY_BUFFER, normalbufferID);
	glBufferData(GL_ARRAY_BUFFER, numNormals*3*sizeof(float), vNormals, GL_STATIC_DRAW);

	if ( numTextureCoords > 0 )
	{
		glGenBuffers(1, &texturebufferID);
		glBindBuffer(GL_ARRAY_BUFFER, texturebufferID);
		glBufferData(GL_ARRAY_BUFFER, numTextureCoords*2*sizeof(float), textureCoords, GL_STATIC_DRAW);
	}

	m_IfVBOReady = true;
}
*/
bool CTriMesh::loadFromObj( const char *filename )
{
	int i;
	strcpy( diskFileName, filename );
    
	GLMmodel *model = glmReadOBJ( filename );
    if (!model)
	{
		wxLogError( "Can not load OBJ file %s.", filename );
		return false;
	}
//    glmUnitize(model);
//    glmFacetNormals(model);
//    glmVertexNormals(model, 90.0);
	numFaces = model->numtriangles;

	numVertice = model->numvertices;
	vertice = (float*)malloc(sizeof(float) * 3 * numVertice);
	memcpy( vertice, model->vertices+3, 3 * numVertice * sizeof(float) );

	numNormals = model->numnormals;
    if (numNormals)
	{
        vNormals = (float*)malloc(sizeof(GLfloat) * 3 * numNormals);
		memcpy( vNormals, model->normals+3, 3 * numNormals * sizeof(float) );
	}

	numTextureCoords = model->numtexcoords;
	if (numTextureCoords) 
	{
        textureCoords = (GLfloat*)malloc(sizeof(GLfloat) * 2 * numTextureCoords  );
		memcpy( textureCoords, model->texcoords+2, 2 * numTextureCoords * sizeof(float) );
	}

	materials.resize( model->nummaterials );
	for ( i = 0; i < materials.size(); i++ )
	{
		materials[i] = new CMaterial( model->materials[i] );
	}

	meshGroups.resize( model->numgroups );
	GLMgroup *group = model->groups;
	for ( i = 0; i < meshGroups.size(); i++ )
	{
		
		meshGroups[i] = new CMeshGroup( this, model, *group );
		group = group->next;
	}
	glmDelete( model );
	updatevfiList();

    if (!numNormals)
		ComputeVNormals();
	return true;
}

void CTriMesh::ComputeVNormals()
{
	float *fp;
	if ( vNormals == NULL )
	{
		numNormals = numVertice;
		vNormals = (float*)malloc(sizeof(float) * 3 * numNormals);
	}
	memset( vNormals, 0, 3*numNormals*sizeof(float) );
	for ( int j = 0; j < numVertice; j++ )
	{
		for ( int k = 0; k < (int) m_vfiList[j].m_faceIndexList.size(); k++ )
		{
			fp = m_vfiList[j].m_meshGroup->fNormals+3*m_vfiList[j].m_faceIndexList[k];
			int jj = 3*j;
			vNormals[jj] += *fp;
			vNormals[jj+1] += *(fp+1);
			vNormals[jj+2] += *(fp+2);
		}
	}
	for ( int i = 0; i < meshGroups.size(); i++ )
		for ( int j = 0; j < meshGroups[i]->numFaces; j++ )
			for ( int k = 0; k < 3; k++ )
				meshGroups[i]->vnIndice[3*j+k] = meshGroups[i]->fIndice[3*j+k];
}

void CTriMesh::updatevfiList()
{
	m_vfiList.clear();
	m_vfiList.resize( numVertice );
	for ( int i = 0; i < meshGroups.size(); i++ )
	{
		for ( int j = 0; j < meshGroups[i]->numFaces; j++ )
		{
			for ( int k = 0; k < 3; k++ )
			{
				m_vfiList[meshGroups[i]->fIndice[3*j+k]].m_meshGroup = meshGroups[i];
				m_vfiList[meshGroups[i]->fIndice[3*j+k]].m_faceIndexList.push_back( j );
			}
		}
	}
}

bool CTriMesh::loadFile( const char *fileName )
{
	char ext[4];
	bool return_value;
	strcpy(ext,fileName+strlen(fileName)-3);
	_strlwr(ext);
	if(strcmp(ext,"obj")==0 || strcmp(ext,"OBJ")==0)
		return_value = loadFromObj( fileName );
	else
	{
		wxLogError( "The mesh format of file %s is not supported.\n", fileName );
		return false;
	}

	computeBoundingBox();
	beginx = beginy = 0.0f;
	ResetPosition();
	trackball( quat, 0.0f, 0.0f, 0.0f, 0.0f );

	return return_value;
}

void CTriMesh::drawWireFrame()
{
	glColor3f( 0.7f, 0.7f, 0.7f );
	glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
	glShadeModel(GL_SMOOTH);
/*
	glEnableClientState(GL_VERTEX_ARRAY);
	glBindBuffer(GL_ARRAY_BUFFER, vertexbufferID);
	glVertexPointer(3,GL_FLOAT,0,0);

	for ( int i = 0; i < meshGroups.size(); i++ )
		glDrawElements(GL_TRIANGLES,3*meshGroups[i]->numFaces,GL_UNSIGNED_INT,meshGroups[i]->fIndice);

	glDisableClientState(GL_VERTEX_ARRAY);
*/

	glBegin( GL_TRIANGLES );
	for ( int i = 0; i < meshGroups.size(); i++ )
	{
		if ( meshGroups[i]->m_visiable )
		{
			for ( int j = 0 ; j < meshGroups[i]->numFaces; j++ )
				for ( int k = 0; k < 3; k++ )
					glVertex3fv( vertice+3*meshGroups[i]->fIndice[3*j+k] );
		}
	}
	glEnd();
}

void CTriMesh::drawGouraud()
{
	glShadeModel(GL_SMOOTH);
	glEnable(GL_LIGHTING);
	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
/*
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	glBindBuffer(GL_ARRAY_BUFFER, vertexbufferID);
	glVertexPointer(3,GL_FLOAT,0,0);
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, normalbufferID);
	glNormalPointer(GL_FLOAT,0,0);

	for ( int i = 0; i < meshGroups.size(); i++ )
		if ( meshGroups[i]->m_visiable )
			glDrawElements(GL_TRIANGLES,3*meshGroups[i]->numFaces,GL_UNSIGNED_INT,meshGroups[i]->fIndice);

	glDisableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);
*/
	glBegin( GL_TRIANGLES );
	for ( int i = 0; i < meshGroups.size(); i++ )
	{
		if ( meshGroups[i]->m_visiable )
		{
			for ( int j = 0 ; j < meshGroups[i]->numFaces; j++ )
				for ( int k = 0; k < 3; k++ )
				{
					glNormal3fv( vNormals+3*meshGroups[i]->vnIndice[3*j+k] );
					glVertex3fv( vertice+3*meshGroups[i]->fIndice[3*j+k] );
				}
		}
	}
	glEnd();

	glDisable(GL_LIGHTING);
}

void CTriMesh::drawTextured()
{
	glShadeModel(GL_SMOOTH);
	glEnable(GL_LIGHTING);
	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );

	for ( int i = 0; i < meshGroups.size(); i++ )
	{
		if ( meshGroups[i]->m_visiable )
		{
			CMaterial *material = materials[meshGroups[i]->materialIndex];
            glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, material->ambient);
            glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, material->diffuse);
            glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, material->specular);
            glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, material->shininess);

			if (material->tex != NULL )
			{
				glEnable( GL_TEXTURE_2D );
				material->tex->MakeCurrent();
				glBegin( GL_TRIANGLES );
				for ( int j = 0 ; j < meshGroups[i]->numFaces; j++ )
				{
					for ( int k = 0; k < 3; k++ )
					{
						glTexCoord2fv( textureCoords+2*meshGroups[i]->vtIndice[3*j+k] );
						glNormal3fv( vNormals+3*meshGroups[i]->vnIndice[3*j+k] );
						glVertex3fv( vertice+3*meshGroups[i]->fIndice[3*j+k] );
					}
				}
				glEnd();
				glDisable( GL_TEXTURE_2D );
			}
			else
			{
				glBegin( GL_TRIANGLES );
				for ( int j = 0 ; j < meshGroups[i]->numFaces; j++ )
				{
					for ( int k = 0; k < 3; k++ )
					{
						glNormal3fv( vNormals+3*meshGroups[i]->vnIndice[3*j+k] );
						glVertex3fv( vertice+3*meshGroups[i]->fIndice[3*j+k] );
					}
				}
				glEnd();
			}
		}
	}
	glDisable(GL_LIGHTING);
}

void CTriMesh::draw()
{
//	if ( !m_IfVBOReady )
//		prepareVBO();

	glPushMatrix();

	glTranslatef( boundingBox.center[0]-position[0], boundingBox.center[1]-position[1], boundingBox.center[2]-position[2] );

	// Transformations
    GLfloat m[4][4];
    build_rotmatrix( m, quat );
    glMultMatrixf( &m[0][0] );

	glTranslatef( -boundingBox.center[0], -boundingBox.center[1], -boundingBox.center[2] );

	switch( m_rendermode )
	{
	case WIREFRAME:
		drawWireFrame();
		break;
	case GOURAUD:
		drawGouraud();
		break;
	case TEXTURED:
		if ( materials.size() != 0 )
			drawTextured();
		else
			m_rendermode = GOURAUD;
		break;
	}
	glPopMatrix();
}

void CTriMesh::destroyMyself()
{
	if ( vertice != NULL ) free( vertice );
	if ( vNormals != NULL ) free ( vNormals );
	if ( textureCoords != NULL ) free( textureCoords );

	for ( int i = 0; i < meshGroups.size(); i++ )
		delete meshGroups[i];
	for ( int i = 0; i < materials.size(); i++ )
		delete materials[i];
}

unsigned int CTriMesh::getFaceNumber()
{
	return numFaces;
}

CMeshGroup::CMeshGroup(CTriMesh *triMesh, GLMmodel *model, GLMgroup &glmg)
: fIndice(NULL), vnIndice(NULL), vtIndice(NULL), fNormals(NULL), m_triMesh(triMesh), m_visiable(true)
{
	strcpy( name, glmg.name );
	materialIndex = glmg.material;
	numFaces = glmg.numtriangles;
	if ( numFaces <= 0 )
		return;
	fIndice = (unsigned int*) malloc( 3 * numFaces * sizeof(unsigned int) );
	vnIndice = (unsigned int*) malloc( 3 * numFaces * sizeof(unsigned int) );
	fNormals = (float*) malloc( 3 * numFaces * sizeof(float) );
	if ( model->numtexcoords )
	{
		vtIndice = (unsigned int*) malloc( 3 * numFaces * sizeof(unsigned int) );
		for ( unsigned int i = 0; i < numFaces; i++ )
			for ( int j = 0; j < 3; j++ )
				vtIndice[3*i+j] = model->triangles[glmg.triangles[i]].tindices[j]-1;
	}
	else
		vtIndice = NULL;
	for ( unsigned int i = 0; i < numFaces; i++ )
	{
		for ( int j = 0; j < 3; j++ )
		{
			fIndice[3*i+j] = model->triangles[glmg.triangles[i]].vindices[j]-1;
			vnIndice[3*i+j] = model->triangles[glmg.triangles[i]].nindices[j]-1;
		}
	}

	ComputeFaceNormals();
}

void CMeshGroup::ComputeFaceNormals()
{
	float v0[3], v1[3];

	for ( int j = 0; j < 3*numFaces; j+=3 )
	{
		glmMinus( m_triMesh->vertice+3*fIndice[j+1], m_triMesh->vertice+3*fIndice[j], v0 );
		glmMinus( m_triMesh->vertice+3*fIndice[j+2], m_triMesh->vertice+3*fIndice[j+1], v1 );
		glmCross( v0, v1, fNormals+j );
		glmNormalize( fNormals+j );
	}
}

CMeshGroup::~CMeshGroup()
{
	if ( fIndice != NULL ) free ( fIndice );
	if ( vtIndice != NULL ) free ( vtIndice );
	if ( vnIndice != NULL ) free ( vnIndice );
	if ( fNormals != NULL ) free (fNormals);
}

#define max(a,b) a < b ? b : a

void CTriMesh::computeBoundingBox()
{
	float mmax[3] = { -9e10, -9e10, -9e10 };
	float mmin[3] = { 9e10, 9e10, 9e10 };

	for ( int i = 0; i < numVertice; i++ )
	{
		for ( int j = 0; j < 3; j++ )
		{
			if ( mmax[j] < vertice[3*i+j] )
				mmax[j] = vertice[3*i+j];
			if ( mmin[j] > vertice[3*i+j] )
				mmin[j] = vertice[3*i+j];
		}
	}
	boundingBox.size = max( max( mmax[0] - mmin[0], mmax[1] - mmin[1] ), mmax[2] - mmin[2] );
	for ( int i = 0 ; i < 3; i++ )
		boundingBox.center[i] = ( mmax[i] + mmin[i] ) * 0.5f;
}

CMaterial::CMaterial(GLMmaterial &glmm)
{
	strcpy( name, glmm.name );
	for ( int i = 0; i < 4; i++ )
	{
		ambient[i] = glmm.ambient[i];
		diffuse[i] = glmm.diffuse[i];
		specular[i] = glmm.specular[i];
		emmissive[i] = glmm.emmissive[i];
	}
	shininess = glmm.shininess;
	tex = glmm.tex;
}


} //end of namespace