#ifndef _TRIMESH_H_
#define _TRIMESH_H_

#include "object.h"
#include "glm.h"
#include "texture.h"

#include <vector>
using namespace std;

namespace AMT
{

class CObject;
class CMeshGroup;
class CTriMesh;

typedef vector<int> TFaceIndexList;

class CMaterial
{
public:
	char name[64];              /* name of material */
	float diffuse[4];           /* diffuse component */
	float ambient[4];           /* ambient component */
	float specular[4];          /* specular component */
	float emmissive[4];         /* emmissive component */
	float shininess;            /* specular exponent */
	Texture* tex;
	CMaterial(GLMmaterial &glmm);
	~CMaterial() { delete tex; }
};

class CMeshGroup
{
public:
	char name[128];
	unsigned int numFaces;
	int materialIndex;

	unsigned int *fIndice;
	unsigned int *vtIndice;
	unsigned int *vnIndice;
	float *fNormals;
	bool m_visiable;

	~CMeshGroup();
	CMeshGroup(CTriMesh *triMesh, GLMmodel *model, GLMgroup &glmg);
	CMeshGroup() : numFaces(0), materialIndex(-1), fIndice(NULL), vtIndice(NULL), \
		vnIndice(NULL), fNormals(NULL), m_triMesh(NULL), m_visiable(true) {};
	void SetTriMesh( CTriMesh *triMesh ){m_triMesh = triMesh; }
private:
	CTriMesh *m_triMesh;
	void ComputeFaceNormals();
};

class CVertexFaceIndex
{
public:
	CMeshGroup *m_meshGroup;
	TFaceIndexList m_faceIndexList;
};

typedef enum {
	WIREFRAME = 0,
	GOURAUD,
	TEXTURED
} TRenderMode;

class CTriMesh : public CRenderObject
{
public:
	CTriMesh();
	CTriMesh( const char *fileName );

	bool loadFile( const char *fileName );

	int numVertice, numFaces, numNormals, numTextureCoords;
	bool m_InverseNormal;

	float *vertice;
	float *vNormals;
	float *textureCoords;

	vector<CMaterial*> materials;
	vector<CVertexFaceIndex> m_vfiList;

	vector<CMeshGroup*> meshGroups;

	virtual void draw();
	virtual void destroyMyself();
	virtual unsigned int getFaceNumber();
	void ChangeRenderMode(TRenderMode rm ) {m_rendermode = rm;}
private:
//	GLuint vertexbufferID, normalbufferID, texturebufferID;
//	bool m_IfVBOReady;

	TRenderMode m_rendermode;
	char diskFileName[128];
	bool loadFromObj( const char *filename );
	void updatevfiList();
	void drawWireFrame();
	void drawGouraud();
	void drawTextured();
	void ComputeVNormals();
//	void prepareVBO();

	void computeBoundingBox();
};

}// end of namespace 

#endif