#ifndef WX_PRECOMP
#include "wx/wx.h"
#endif

#include <GL/glew.h>
#include "renderFrame.h"
#include "object.h"
#include "trimesh.h"
#include "gllight.h"
#include "GL/glu.h"
#include "GL/glut.h"

#define ONE_SECOND 1
extern "C"
{
#include "trackball.h"
}

#define FIELD_OF_VIEW 15.0

namespace AMT
{

vector<CRenderFrame*> CRenderFrame::renderFrameList;
unsigned int CRenderFrame::maxNumFrames = 0;
CRenderFrame* CRenderFrame::lastFocusedFrame = NULL;

void render_string(void* font, const char* string)
{
	char* p = (char*) string;
	while (*p != '\0') glutBitmapCharacter(font, *p++);
}

enum
{
    ID_QUIT  = wxID_EXIT,
    ID_NEW = 100
};

BEGIN_EVENT_TABLE(CRenderFrame, wxFrame)
    EVT_MENU(ID_QUIT, CRenderFrame::OnExit)
    EVT_MENU(ID_NEW, CRenderFrame::OnNewFrame)
//	EVT_SET_FOCUS(CRenderFrame::OnFocusGot)
END_EVENT_TABLE()

CRenderFrame *CRenderFrame::Create( const wxString& title, const wxPoint& pos,
        const wxSize& size, long style )
{
	CRenderFrame *rf = new CRenderFrame(title, pos, size, style);
	
	rf->SetCanvas( new CGLCanvas( rf, wxID_ANY,wxDefaultPosition, size) );

	renderFrameList.push_back( rf );
	maxNumFrames++;

    rf->Show(true);

    return rf;
}

CRenderFrame::~CRenderFrame()
{
	if ( lastFocusedFrame == this )
		lastFocusedFrame = NULL;
	for ( int i = 0; i < renderFrameList.size(); i++ )
	{
		if ( renderFrameList[i] == this )
		{
			renderFrameList.erase(renderFrameList.begin() + i);
			break;
		}
	}
	if ( m_canvas != NULL )
		delete m_canvas;

	for ( int i = 0; i < m_renderObjectList.size(); i++ )
		m_renderObjectList[i]->destroyMyself();
}
/*
void CRenderFrame::OnFocusGot(wxFocusEvent& event)
{
	CRenderFrame::lastFocusedFrame = this;
}
*/
void CRenderFrame::AddRenderObject( CRenderObject *ro )
{
	m_renderObjectList.push_back( ro );
	m_canvas->numTriangleCount += ro->getFaceNumber();
}

void CRenderFrame::OnNewFrame(wxCommandEvent& event)
{
	char temp[32];
	sprintf( temp, "Render Window #%d", maxNumFrames );
	(void) CRenderFrame::Create(temp, wxDefaultPosition, wxSize(800, 800));
}

CRenderFrame::CRenderFrame(const wxString& title, const wxPoint& pos,
						   const wxSize& size, long style)
						   : wxFrame( NULL, wxID_ANY, title, pos, size, style ), selectedObject(NULL)
{
    wxMenu *fileMenu = new wxMenu;

    fileMenu->Append(ID_NEW, wxT("&New Render Window"));
    fileMenu->Append(ID_QUIT, wxT("E&xit"));
    wxMenuBar *menuBar = new wxMenuBar;
    menuBar->Append(fileMenu, wxT("&File"));
    SetMenuBar(menuBar);
}

void CRenderFrame::OnExit( wxCommandEvent& WXUNUSED(event) )
{
    Close(true);
}

BEGIN_EVENT_TABLE(CGLCanvas, wxGLCanvas)
    EVT_SIZE(CGLCanvas::OnSize)
    EVT_PAINT(CGLCanvas::OnPaint)
    EVT_ERASE_BACKGROUND(CGLCanvas::OnEraseBackground)
	EVT_SET_FOCUS(CGLCanvas::OnFocusGot)
	EVT_KILL_FOCUS(CGLCanvas::OnFocusKill)
	EVT_KEY_DOWN(CGLCanvas::OnKeyDown)
    EVT_MOUSE_EVENTS(CGLCanvas::OnMouse)
	EVT_IDLE(CGLCanvas::OnIdle) 
END_EVENT_TABLE()

CGLCanvas::CGLCanvas( wxWindow *parent, wxWindowID id,
    const wxPoint& pos, const wxSize& size, long style, const wxString& name)
    : wxGLCanvas(parent, (wxGLCanvas*) NULL, id, pos, size, style|wxFULL_REPAINT_ON_RESIZE , name )
{
	rf = (CRenderFrame*) parent;
    m_init = false;
	m_fieldofview = FIELD_OF_VIEW;
	m_size = size;
	m_viewPoint[0] = 0.0f; m_viewPoint[1] = 0.0f; m_viewPoint[2] = 15.0f;
	m_viewCenter[0] = 0.0f; m_viewCenter[1] = 0.0f; m_viewCenter[2] = 0.0f;
	m_upVector[0] = 0.0f; m_upVector[1] = 1.0f; m_upVector[2] = 0.0f;
	xTrans = yTrans = zScale = 0.0f;

	lastTime = 0;
	numTriangleCount = 0;
}

CGLCanvas::~CGLCanvas()
{
	for ( int i = 0; i < glLightList.size(); i++ )
		delete glLightList[i];
}

void CGLCanvas::InitGL()
{
	SetCurrent();
	GLenum err = glewInit();
	if (GLEW_OK != err)
		wxLogError( "%s\n", glewGetErrorString(err) );

//	if ( GLEW_ARB_vertex_buffer_object )
//		wxLogMessage( "VBO extersion is supported!\n");


	glClearColor( 0.15f, 0.2f, 0.3f, 1.0f );
	glClearDepth(1.0);

    glEnable(GL_LINE_SMOOTH);
    glEnable(GL_BLEND);

    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glHint(GL_LINE_SMOOTH_HINT, GL_DONT_CARE);
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_DEPTH_TEST);
 	glEnable(GL_NORMALIZE);

	glview_text.Load("font.tga"); 

	CGLLight *gll = CGLLight::CreateLight();
	glLightList.push_back( gll );
}

void CGLCanvas::DrawTriangleCount()
{
	char str[64];
	sprintf( str, "Number of Triangles: %d", numTriangleCount );
	glview_text.Size(12);
	glview_text.Draw( 5, 25, str );
}

void CGLCanvas::Render()
{
    wxPaintDC dc(this);
	if ( !GetContext() )
		return;
	SetCurrent();
    // Init OpenGL once, but after SetCurrent
    if (!m_init)
    {
        InitGL();
        m_init = true;
    }

	int w = m_size.x, h = m_size.y;
	glViewport (0, 0, w, h );
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	gluPerspective(m_fieldofview, (float) w / h, .1, 1000.0);
	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	gluLookAt( m_viewPoint[0], m_viewPoint[1], m_viewPoint[2],
			   m_viewCenter[0], m_viewCenter[1], m_viewCenter[2],
			   m_upVector[0], m_upVector[1], m_upVector[2] );

	glLightList[0]->SetPosition( m_viewPoint[0], m_viewPoint[1], m_viewPoint[2], 1.0f );
	glLightList[0]->TurnOn();

	for ( int i = 0; i < rf->m_renderObjectList.size(); i++ )
	{
		if ( rf->m_renderObjectList[i]->m_visiable )
			rf->m_renderObjectList[i]->draw();
	}
	if ( CRenderFrame::lastFocusedFrame == rf && numTriangleCount > 0 )
		CalculateFrameRate();

	if ( numTriangleCount > 0 )
		DrawTriangleCount();

    glFlush();
    SwapBuffers();
}

void CGLCanvas::OnFocusGot( wxFocusEvent& event )
{
//	SetCurrent();
//	wxLogMessage("CGLCanvas::FocusGot has been called.");
	CRenderFrame::lastFocusedFrame = rf;
	SetExtraStyle( wxWS_EX_PROCESS_IDLE );
}

void CGLCanvas::OnFocusKill( wxFocusEvent& event )
{
//	SetCurrent();
//	wxLogMessage("CGLCanvas::FocusGot has been called.");
	SetExtraStyle( 0 );
	Refresh(false);
}

void CGLCanvas::OnIdle( wxIdleEvent& event )
{
	Refresh(false);
}

void CGLCanvas::OnPaint( wxPaintEvent& WXUNUSED(event) )
{
//	wxLogMessage("CGLCanvas::OnPaint has been called.");
    Render();
}

void CGLCanvas::OnSize(wxSizeEvent& event)
{
    // this is also necessary to update the context on some platforms
//    wxGLCanvas::OnSize(event);
    // set GL viewport (not called by wxGLCanvas::OnSize on all platforms...)
    int w, h;
    GetClientSize(&w, &h);

	glViewport (0, 0, w, h ); 
    m_size.SetHeight( h );
	m_size.SetWidth( w );
}

void CGLCanvas::OnEraseBackground(wxEraseEvent& WXUNUSED(event))
{
  // Do nothing, to avoid flashing.
}

void CGLCanvas::OnKeyDown( wxKeyEvent& event )
{
	SetCurrent();
	char ch = event.GetKeyCode();
	if (  ch == 'f' || ch == 'F' )
		rePositionCamera();
	else if ( ch == '4' )
	{
		if ( rf->selectedObject->m_IsTriMesh )
		{
			CTriMesh *tm = (CTriMesh*) rf->selectedObject;
			tm->ChangeRenderMode( WIREFRAME );
		}
	}
	else if ( ch == '5' )
	{
		if ( rf->selectedObject->m_IsTriMesh )
		{
			CTriMesh *tm = (CTriMesh*) rf->selectedObject;
			tm->ChangeRenderMode( GOURAUD );
		}
	}
	else if ( ch == '6' )
	{
		if ( rf->selectedObject->m_IsTriMesh )
		{
			CTriMesh *tm = (CTriMesh*) rf->selectedObject;
			tm->ChangeRenderMode( TEXTURED );
		}
	}
	else if ( ch == 'i' || ch == 'I' )
	{
		CTriMesh *tm = (CTriMesh*) rf->selectedObject;
		tm->m_InverseNormal = ! tm->m_InverseNormal;
	}
	Refresh(false);
}

void CGLCanvas::rePositionCamera()
{
	CRenderFrame *rf = (CRenderFrame*) this->GetParent();
	CBoundingBox bb;

	if ( rf->selectedObject != NULL )
	{
		bb = rf->selectedObject->boundingBox;
		rf->selectedObject->ResetPosition();
	}
	else
	{
		// Go through every object and find the common bounding box
	}

	float temp = bb.size * 0.5f / tan( (m_fieldofview - 8.0f) * 3.14159265 / 180.0f ) + bb.size * 0.5f;

//	m_viewCenter[0] = bb.center[0]; m_viewCenter[1] = bb.center[1]; m_viewCenter[2] = bb.center[2];
//	m_viewPoint[0] = bb.center[0]; m_viewPoint[1] = bb.center[1]; m_viewPoint[2] = bb.center[2] + temp; 
	m_viewPoint[2] = bb.center[2] + temp;
}

void CGLCanvas::OnMouse( wxMouseEvent& event )
{
	CRenderObject *ro = rf->selectedObject;
	if ( ro == NULL )
		return;
	int mouse_x = event.GetX();
	int mouse_y = event.GetY();

	if ( event.AltDown() )
	{
		if ( event.ButtonDown() )
		{
			pointx = mouse_x;
			pointy = mouse_y;
		}
		else if ( event.ButtonUp() )
		{
			xTrans = ro->position[0];
			yTrans = ro->position[1];
			zScale = ro->position[2];
		}
		else if ( event.Dragging() )
		{
			wxSize sz( GetClientSize() );

			if ( event.LeftIsDown() )
			{
				/* drag in progress, simulate trackball */
				float spin_quat[4];
				trackball(spin_quat,
					(2.0*ro->beginx -       sz.x) / sz.x,
					(     sz.y - 2.0*ro->beginy) / sz.y,
					(     2.0*event.GetX() - sz.x) / sz.x,
					(    sz.y - 2.0*event.GetY()) / sz.y);

				add_quats( spin_quat, ro->quat, ro->quat );
				/* orientation has changed, redraw mesh */
				Refresh(false);
			}
			else if ( event.RightIsDown() )
			{
				ro->position[0] = xTrans - (float)( mouse_x - pointx) / 5.0f;
				ro->position[1] = yTrans + (float)( mouse_y - pointy) / 5.0f;
			}
			else if ( event.MiddleIsDown() )
			{
				ro->position[2] = zScale + (float)( pointy - mouse_y ) / 5.0f;
			}

		}
	}

    ro->beginx = mouse_x;
    ro->beginy = mouse_y;
}

void CGLCanvas::CalculateFrameRate()
{
	SetCurrent();
    long cur_t = (long)((long)(clock()) / 1000);

    // Increase the frame counter
    ++framesPerSecond;

    if( cur_t - lastTime >= ONE_SECOND )
    {
        lastTime = cur_t;

        // Copy the frames per second into a string to display in the window
        sprintf(strFrameRate, "Frames Rate: %d", framesPerSecond);

        // Reset the frames per second
        framesPerSecond = 0;
    }

    // draw frame rate on screen
	glview_text.Size(12);
	glview_text.Draw( 5, 10, strFrameRate );
}


} // end of namespace