#ifndef _RENDERFRAME_H_
#define _RENDERFRAME_H_

#include <vector>
using namespace std;
#include "wx/glcanvas.h"
#include "text.h"

namespace AMT
{

class CGLCanvas;
class CRenderObject;
class CRenderFrame;
class CControlFrame;
class CGLLight;

class CRenderFrame: public wxFrame
{
	friend class CGLCanvas;
	friend class CControlFrame;
public:
	static unsigned int maxNumFrames;
	static vector<CRenderFrame*> renderFrameList;
	static CRenderFrame* lastFocusedFrame;
	static CRenderFrame *Create( const wxString& title, const wxPoint& pos,
        const wxSize& size, long style = wxDEFAULT_FRAME_STYLE );

    CRenderFrame( const wxString& title, const wxPoint& pos,
        const wxSize& size, long style = wxDEFAULT_FRAME_STYLE);
	~CRenderFrame();

    void OnExit(wxCommandEvent& event);
//	void OnFocusGot(wxFocusEvent& event);

    void SetCanvas( CGLCanvas *canvas ) { m_canvas = canvas; }
    CGLCanvas *GetCanvas() { return m_canvas; }

	void OnNewFrame(wxCommandEvent& event);
	void AddRenderObject( CRenderObject *ro );

private:
    CGLCanvas *m_canvas;
	vector<CRenderObject*> m_renderObjectList;
	CRenderObject *selectedObject;

    DECLARE_EVENT_TABLE()
};

class CGLCanvas: public wxGLCanvas
{
    friend class CRenderFrame;
public:
	vector<CGLLight*> glLightList; 
    CGLCanvas( wxWindow *parent, wxWindowID id = wxID_ANY,
        const wxPoint& pos = wxDefaultPosition,
        const wxSize& size = wxDefaultSize,
        long style = 0, const wxString& name = _T("CGLCanvas") );

 	~CGLCanvas();

    void OnPaint(wxPaintEvent& event);
    void OnSize(wxSizeEvent& event);
    void OnEraseBackground(wxEraseEvent& event);
    void OnFocusGot(wxFocusEvent& event);
    void OnFocusKill(wxFocusEvent& event);
	void OnKeyDown( wxKeyEvent& event );
	void OnMouse( wxMouseEvent& event );
	void OnIdle( wxIdleEvent& event );

    void Render();
    void InitGL();

private:

	CRenderFrame *rf;
	bool m_init;
	wxSize m_size;
	float m_fieldofview;
	float m_viewPoint[3];
	float m_viewCenter[3];
	float m_upVector[3];

	float pointx, pointy, xTrans, yTrans, zScale;

	float quat[4];        /* orientation of object */

    int framesPerSecond;        // This will store our fps
    long lastTime;           // This will hold the time from the last frame
    char strFrameRate[50];         // We will store the string here for the window title

	unsigned int numTriangleCount;
	
	text_t glview_text;

	void CalculateFrameRate();
	void rePositionCamera();
	void DrawTriangleCount();

	DECLARE_EVENT_TABLE()
};

}
#endif